from __future__ import annotations

from os import PathLike
from pathlib import Path
import re
import numpy

from .file import read
from .header import Header


def _get_channel_trace_from_trc_filename(
    filename: str | PathLike[str],
) -> tuple[int, int] | None:
    """
    Get channel number from trc filename.
    trc files follow the pattern 'C{n}Trace{NNNNNN}.trc' where n is the channel number and NNNNNN is the trace number.
    If the filename does not follow this pattern, None is returned.
    """

    # get filename from path
    filename = Path(filename).name

    regex = re.compile(r"C(\d+)Trace(\d\d\d\d\d).trc")
    match = regex.match(filename)

    if match is None:
        return None

    channel = int(match.group(1))
    trace_number = int(match.group(2).lstrip("0"))
    return channel, trace_number


class Trace:
    def __init__(
        self,
        filename_or_bytes: str | PathLike[str] | bytes,
        header_only: bool = False,
        channel: int | None = None,
    ):
        self._filename = (
            filename_or_bytes if not isinstance(filename_or_bytes, bytes) else ""
        )

        self._channel = None
        if channel is not None:
            self.channel = channel
        else:
            if isinstance(filename_or_bytes, bytes):
                try:
                    channel_string = filename_or_bytes[0:5].decode("ascii")
                    regex = re.compile(r"C(\d+):WF")
                    match = regex.match(channel_string)
                    if match:
                        self.channel = int(match.group(1))
                except UnicodeDecodeError:
                    pass
            else:
                # attempt to get channel number from filename
                channel_trace = _get_channel_trace_from_trc_filename(filename_or_bytes)
                if channel_trace is not None:
                    self.channel = channel_trace[0]

        header, self._trigger_times, self._voltage = read(
            filename_or_bytes, header_only
        )
        self._header = Header(header)

        # store values in voltage units
        self._voltage = self._voltage * self.header["vertical_gain"]
        self._voltage = self._voltage - self.header["vertical_offset"]

        # compute time values
        self._time = (
            numpy.arange(self._voltage.shape[-1]) * self.header["horiz_interval"]
            + self.header["horiz_offset"]
        )

    def __len__(self):
        if not self._header.sequence:
            return 1
        return len(self._voltage)

    def __iter__(self):
        if len(self) == 1:
            yield self._time, self._voltage
        else:
            for single in self._voltage:
                yield self._time, single

    @property
    def header(self) -> Header:
        return self._header

    @property
    def header_only(self) -> bool:
        return self._voltage.size == 0

    @property
    def sequence(self) -> bool:
        return self.header.sequence

    @property
    def channel(self) -> int:
        return self._channel

    @channel.setter
    def channel(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Channel number must be a positive integer (1, 2, 3, ...)")
        self._channel = value

    @property
    def voltage(self) -> numpy.ndarray:
        return self._voltage

    @property
    def adc_values(self) -> numpy.ndarray:
        return numpy.array(
            (self._voltage + self.header["vertical_offset"])
            / self.header["vertical_gain"],
            dtype=int,
        )

    @property
    def trigger_times(self) -> numpy.ndarray:
        return self._trigger_times

    @property
    def time(self) -> numpy.ndarray:
        return self._time

    # Alternative names
    @property
    def x(self) -> numpy.ndarray:
        return self._time

    @property
    def y(self) -> numpy.ndarray:
        return self._voltage



from __future__ import annotations
from functools import cached_property

import vxi11
import ipaddress

from lecroyscope import Trace, TraceGroup


# The following manual can be references for additional commands and documentation:
# https://cdn.teledynelecroy.com/files/manuals/automation_command_ref_manual_wm-wp.pdf


def _parse_response(response: str) -> str:
    if response.startswith("VBS "):
        response = response[4:]
    return response.strip()


def _set_command(
    parameter: str, value: str | int | float, return_value: bool = False
) -> str:
    return_command = f"return = app.{parameter}" if return_value else ""
    return f"""VBS? \'app.{parameter} = "{value}"\n{return_command}\'"""


def _get_command(parameter: str) -> str:
    return f"VBS? 'return = app.{parameter}'"


def _capitalize_first_letter(string: str) -> str:
    if len(string) == 0:
        return string
    return string[0].upper() + string.lower()[1:]


class Scope:
    def __init__(self, ip_address: str):
        # validate ip address
        ip_address = str(ipaddress.ip_address(ip_address))

        self._instrument = vxi11.Instrument(ip_address)
        self.timeout = 30.0

    def __del__(self):
        self._instrument.close()

    @property
    def timeout(self) -> float:
        """ " Returns instrument timeout in seconds"""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: float):
        """ " Sets instrument timeout in seconds"""
        if timeout <= 0.0:
            raise ValueError("Scope timeout must be a positive floating point number")
        self._timeout = timeout
        self._instrument.timeout = self._timeout

    def _ask(self, command: str):
        return _parse_response(self.instrument.ask(command))

    def get(self, parameter: str):
        return self._ask(_get_command(parameter))

    def set(self, parameter: str, value: str | int | float):
        return self._ask(_set_command(parameter, value, return_value=True))

    def acquire(self, force: bool = False) -> bool:
        response = self._ask(
            f"VBS? 'return = app.Acquisition.acquire({self.timeout}, {force})'"
        )
        if response not in ["0", "1"]:
            raise ValueError(f"Unexpected response from scope: {response}")
        return bool(int(response))

    def read(self, *channels: int) -> Trace | TraceGroup:
        if len(channels) == 0:
            raise ValueError("At least one channel must be specified")
        traces = []
        for channel in channels:
            self.instrument.write(f"C{channel}:WF?")
            trace = Trace(self.instrument.read_raw())
            if trace.channel != channel:
                raise ValueError(f"Unexpected channel: {trace.channel}")
            traces.append(trace)
        return traces[0] if len(traces) == 1 else TraceGroup(*traces)

    @property
    def instrument(self):
        return self._instrument

    @cached_property
    def id(self):
        return self.instrument.ask("*IDN?")

    @cached_property
    def name_all(self):
        name_all = self._ask(_get_command("ExecsNameAll"))
        return name_all.split(",")

    @property
    def sample_mode(self):
        return self._ask(_get_command("Acquisition.Horizontal.SampleMode"))

    @sample_mode.setter
    def sample_mode(self, value: str):
        self._ask(_set_command("Acquisition.Horizontal.SampleMode", value))

    @property
    def num_segments(self):
        return int(self._ask(_get_command("Acquisition.Horizontal.NumSegments")))

    @num_segments.setter
    def num_segments(self, value: int):
        self._ask(_set_command("Acquisition.Horizontal.NumSegments", value))

    @property
    def trigger_mode(self):
        return self._ask(_get_command("Acquisition.TriggerMode"))

    @trigger_mode.setter
    def trigger_mode(self, value: str):
        trigger_modes = ["Stopped", "Single", "Normal", "Auto"]
        if _capitalize_first_letter(value) not in trigger_modes:
            raise ValueError(
                f"Invalid trigger mode: {value}. Valid trigger modes are: {trigger_modes}"
            )
        self._ask(_set_command("Acquisition.TriggerMode", value))

    def channel(self, channel: int) -> Channel:
        return Channel(scope=self, channel=channel)


class Channel:
    def __init__(self, scope: Scope, channel: int):
        self._scope = scope
        self.channel = channel

    def __str__(self):
        return f"C{self._channel:d}"

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value: int):
        if f"C{value:d}" not in self._scope.name_all:
            valid_channels = [
                c for c in self._scope.name_all if c.startswith("C") and len(c) == 2
            ]
            raise ValueError(
                f"""Invalid channel: {value}. Valid channels are: {valid_channels}"""
            )
        self._channel = value

    @property
    def vertical_scale(self) -> float:
        return float(
            self._scope._ask(_get_command(f"Acquisition.{str(self)}.VerScale"))
        )

    @vertical_scale.setter
    def vertical_scale(self, value: float) -> None:
        self._scope._ask(_set_command(f"Acquisition.{str(self)}.VerScale", value))

    @property
    def vertical_offset(self) -> float:
        return float(
            self._scope._ask(_get_command(f"Acquisition.{str(self)}.VerOffset"))
        )

    @vertical_offset.setter
    def vertical_offset(self, value: float) -> None:
        self._scope._ask(_set_command(f"Acquisition.{str(self)}.VerOffset", value))

    # TODO: add setter
    @property
    def vertical_coupling(self) -> str:
        return self._scope._ask(_get_command(f"Acquisition.{str(self)}.Coupling"))

    def find_scale(self) -> None:
        self._scope._ask(f"VBS? 'app.Acquisition.{str(self)}.FindScale()'")

    def read(self) -> Trace:
        return self._scope.read(self.channel)



import pytest

from pathlib import Path
import numpy as np
import lecroyscope

files_path = Path(__file__).parent / "files"


def test_trace_group_from_files(tmp_path):
    filenames = []
    channels = [1, 2, 3, 4]
    for channel in channels:
        filename = f"C{channel}Trace00001.trc"
        tmp_file = tmp_path / filename
        tmp_file.write_bytes((files_path / "header.trc").read_bytes())
        filenames.append(tmp_file)

    # check glob works
    for trace_group in [
        lecroyscope.TraceGroup(*filenames),
        lecroyscope.TraceGroup(tmp_path / "C*Trace00001.trc"),
    ]:

        for i, trace in enumerate(trace_group):
            assert isinstance(trace, lecroyscope.Trace)
            # this checks sorting too! (glob order is not the same across platforms)
            assert trace.channel == channels[i]
            np.testing.assert_array_equal(trace.time, trace_group.time)

        assert len(trace_group) == len(channels)

        for channel in channels:
            assert channel == trace_group[channel].channel



import pytest
from pathlib import Path
import lecroyscope
from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy as np

uproot = pytest.importorskip("uproot")

files_path = Path(__file__).parent / "files"


def test_writing_tree(tmp_path):
    for i, (file, multiply) in enumerate(
        zip(
            [
                files_path / "pulse.trc",
                files_path / "pulse.trc",
                files_path / "pulse_sequence.trc",
                files_path / "pulse_sequence.trc",
                files_path / "pulse_sequence.trc",
            ],
            [False, False, False, True, False],
        )
    ):
        if not multiply:
            trace_group = lecroyscope.TraceGroup(lecroyscope.Trace(file, channel=2))
        else:
            trace_group = lecroyscope.TraceGroup(
                *[lecroyscope.Trace(file, channel=i + 1) for i in range(4)]
            )

        branches = lecroyscope.writing.root.get_tree_branch_definitions(trace_group)
        if not multiply:
            # get_tree_branch_definitions supports Trace or TraceGroup
            assert branches == lecroyscope.writing.root.get_tree_branch_definitions(
                next(trace_group)
            )

        use_additional_info = i == 1 or i == 3
        if use_additional_info:
            branches = lecroyscope.writing.root.get_tree_branch_definitions(
                trace_group, additional_float=float, additional_int=int
            )
            additional_float = 1.25
            additional_int = 137

        tree_name = "t"
        uproot_file = tmp_path / "pulse.root"
        with uproot.recreate(uproot_file) as f:
            tree = f.mktree(
                tree_name,
                branches,
                "Tree",
            )
            data = lecroyscope.writing.root.get_tree_extend_data(trace_group)
            if use_additional_info:
                data = lecroyscope.writing.root.get_tree_extend_data(
                    trace_group,
                    additional_float=additional_float,
                    additional_int=additional_int,
                )
            tree.extend(data)

        with uproot.open(uproot_file) as f:
            tree = f[tree_name]
            times = tree["time"].array()
            voltages = {
                channel: tree[f"CH{channel:d}"].array()
                for channel in trace_group.channels
            }
            if use_additional_info:
                additional_float_file = tree["additional_float"].array()
                additional_int_file = tree["additional_int"].array()
                assert_array_equal(
                    additional_float_file, len(times) * [additional_float]
                )
                assert_array_almost_equal(
                    additional_int_file, len(times) * [additional_int]
                )

        # check all times are equal
        for time in times:
            assert_array_equal(time, times[0])

        # not exactly equal
        assert_array_almost_equal(times[0], trace_group.time)

        for channel in trace_group.channels:
            trace = trace_group[channel]
            assert_array_almost_equal(
                np.array(voltages[channel]), trace.voltage.reshape((len(trace), -1))
            )



import pytest

from numpy.testing import assert_array_equal
import numpy as np
from pathlib import Path

import lecroyscope

from .test_header import header_reference_dict

files_path = Path(__file__).parent / "files"


def test_trace_header():
    filename = files_path / "header.trc"
    trace = lecroyscope.Trace(filename, header_only=True)
    assert dict(trace.header) == header_reference_dict


def test_trace_iter():
    filename = files_path / "header.trc"
    trace = lecroyscope.Trace(filename)

    # this header file is from a sequence of size 200
    assert len(trace) == 200

    data = np.array([(time, single) for time, single in trace])
    assert len(data) == len(trace)
    times = data[:, 0]
    voltage = data[:, 1]
    assert len(times) == len(voltage) == len(trace)

    # check all times are equal
    for time in times:
        assert_array_equal(time, times[0])


def test_trace_channel():
    filename = files_path / "header.trc"

    trace = lecroyscope.Trace(filename)
    assert trace.channel is None

    trace.channel = 2
    assert trace.channel == 2

    with pytest.raises(ValueError):
        trace.channel = 0

    trace = lecroyscope.Trace(filename, channel=3)
    assert trace.channel == 3


def test_trace_channel_from_filename_helper():
    assert (
        lecroyscope.reading._get_channel_trace_from_trc_filename("invalid.trc") is None
    )
    assert lecroyscope.reading._get_channel_trace_from_trc_filename(
        "C2Trace00001.trc"
    ) == (2, 1)
    assert lecroyscope.reading._get_channel_trace_from_trc_filename(
        "C3Trace01021.trc"
    ) == (3, 1021)
    assert lecroyscope.reading._get_channel_trace_from_trc_filename(
        "/this/is/ignored/C2Trace00001.trc"
    ) == (2, 1)
    assert (
        lecroyscope.reading._get_channel_trace_from_trc_filename("C3Tra1ce01021.trc")
        is None
    )
    assert (
        lecroyscope.reading._get_channel_trace_from_trc_filename("C3Trace01.trc")
        is None
    )


def test_trace_channel_from_filename(tmp_path):
    for filename, channel in [
        ("C2Trace00001.trc", 2),
        ("C3Trace01021.trc", 3),
        ("C4Trace01241.trc", 4),
        ("none.trc", None),
    ]:
        tmp_file = tmp_path / filename
        tmp_file.write_bytes((files_path / "header.trc").read_bytes())

        trace = lecroyscope.Trace(tmp_file)
        assert trace.channel == channel


def test_read_trace_from_file():
    for filename, shape, length, sequence in zip(
        [files_path / "pulse.trc", files_path / "pulse_sequence.trc"],
        [(251,), (20, 251)],
        [1, 20],
        [False, True],
    ):
        trace = lecroyscope.Trace(filename)
        assert len(trace) == length
        assert trace.header_only is False
        assert trace.sequence == sequence
        assert trace.voltage.shape == shape
        assert trace.time.shape == (251,)
        assert (trace.voltage.shape[-1],) == trace.time.shape

        if not sequence:
            # check voltage and time scaling is done properly
            # test against signal which consists of ~ 0V baseline + sharp pulse at trigger time
            # first bins should be ~ 0V since we are on the region before pulse
            assert pytest.approx(np.mean(trace.voltage[0:100])) == 0.0057997689768671985
            # time of max voltage value should be very close to trigger time (0 seconds)
            assert (
                pytest.approx(trace.time[np.argmax(trace.y)]) == 4.254989846811945e-09
            )


def test_trace_adc_values():
    filename = files_path / "pulse.trc"
    trace = lecroyscope.Trace(filename)

    adc_from_trace = trace.adc_values

    (
        _,
        _,
        adc_from_file,
    ) = lecroyscope.reading.read(filename)

    assert_array_equal(adc_from_file, adc_from_trace)



from __future__ import annotations

import uproot
import awkward
import numpy
from lecroyscope import Trace, TraceGroup


def get_tree_branch_definitions(trace_or_group: Trace | TraceGroup, **kwargs) -> dict:
    if isinstance(trace_or_group, TraceGroup):
        trace_group = trace_or_group
    elif isinstance(trace_or_group, Trace):
        trace_group = TraceGroup(trace_or_group)
    else:
        raise ValueError(
            "Input argument should be either 'lecroyscope.Trace' or 'lecroyscope.TraceGroup'"
        )
    if trace_group.time is None:
        raise ValueError(
            "Trace Group does not have unified time information (probably do not come from the same trigger)"
        )
    channels = {
        f"CH{trace.channel}": ("f4", (len(trace_group.time),)) for trace in trace_group
    }
    branches = {
        "time": ("f4", (len(trace_group.time),)),
        **channels,
        **kwargs,
    }

    return branches


def get_tree_extend_data(trace_or_group: Trace | TraceGroup, **kwargs):
    """
    Returns a dict which can be used as argument for the uproot tree extend method.
    The resulting dict will have one "time" entry and one channel entry "CHx" per channel with corresponding data
    Optionally kwargs can be used to set "additional" data which must be specified if it was also specified in the
    "get_tree_branch_definitions" method.

    For example:
    If we call `get_tree_branch_definitions(trace, some_data=float)`,
    then to get tree extend data we must call `get_tree_extend_data(trace, some_data=1.25)`.
    The value will be the same for all sequences of the trace (if is sequence).
    If you want different data for each sequence, you can always extend the result of this method invoked without kwargs
    """
    if isinstance(trace_or_group, TraceGroup):
        trace_group = trace_or_group
    elif isinstance(trace_or_group, Trace):
        trace_group = TraceGroup(trace_or_group)
    else:
        raise ValueError(
            "Input argument should be either 'lecroyscope.Trace' or 'lecroyscope.TraceGroup'"
        )

    channels = {
        f"CH{trace.channel}": trace.voltage.reshape((trace_group.trace_length, -1))
        for trace in trace_group
    }

    kwargs = {
        key: numpy.array(trace_group.trace_length * [value])
        for key, value in kwargs.items()
    }

    return {
        "time": numpy.array(trace_group.trace_length * [trace_group.time]).reshape(
            (trace_group.trace_length, -1)
        ),
        **channels,
        **kwargs,
    }



from __future__ import annotations

from os import PathLike

import numpy

from pathlib import Path
from glob import glob

from .trace import Trace


class TraceGroup:
    def __init__(self, *args: str | PathLike[str] | Trace | bytes) -> None:
        self._traces = dict()
        for arg in args:
            traces = []
            if isinstance(arg, Trace):
                traces = [arg]
            elif isinstance(arg, bytes):
                trace = Trace(arg)
                if trace.channel is None:
                    raise ValueError(
                        "Trace group cannot be constructed from bytes without channel number"
                    )
                traces.append(trace)
            else:
                # is pathlike string
                path = Path(arg)
                if "*" in str(path):
                    traces = [Trace(filename) for filename in glob(str(path))]
                else:
                    traces = [Trace(arg)]

            for trace in traces:
                if trace.channel is None:
                    raise ValueError("Trace must have a channel number")

                if trace.channel in self._traces:
                    raise ValueError(
                        f"Channel {trace.channel} already exists in trace group"
                    )

                self._traces[trace.channel] = trace

        # sort by channel number
        self._traces = dict(sorted(self._traces.items()))

    def __iter__(self):
        for trace in self._traces.values():
            yield trace

    def __next__(self):
        return next(iter(self._traces.values()))

    def __len__(self) -> int:
        return len(self._traces)

    def __getitem__(self, item: int | slice) -> TraceGroup | Trace:
        if isinstance(item, slice):
            # when slicing, return do not use channel number as key
            return TraceGroup(*list(self._traces.values())[item])
        if item not in self._traces:
            raise KeyError(f"Trace group does not contain channel {item}")
        return self._traces[item]

    @property
    def channels(self) -> list[int]:
        return list(self._traces.keys())

    @property
    def trace_length(self) -> int | None:
        """
        Returns length of the traces if they are all equal, otherwise return None
        """
        length_reference = len(self[self.channels[0]])
        for trace in self[1:]:
            if len(trace) != length_reference:
                return None
        return length_reference

    @property
    def time(self) -> numpy.ndarray | None:
        """
        Returns the time vector of the traces if they are all equal, otherwise return None
        """
        time_reference = self[self.channels[0]].time
        for trace in self[1:]:
            diff = numpy.abs(trace.time - time_reference)
            if numpy.any(diff > 1e-12):
                return None
        return time_reference

    @property
    def x(self) -> numpy.ndarray | None:
        return self.time
